import asyncio

from aiohttp import web, http_websocket
from nacl.exceptions import BadSignatureError
from nacl.signing import VerifyKey

from v6d0auth import certs, cdb

__all__ = ('V6D0AuthAppFactory',)

from v6d0auth.appfactory import AppFactory


class V6D0AuthAppFactory(AppFactory):
    def __init__(self, loop: asyncio.AbstractEventLoop):
        self.loop = loop

    def define_routes(self, routes: web.RouteTableDef):
        print(certs.vkey.encode().hex())
        mycdb = cdb.CDB(self.loop)
        self.loop.create_task(mycdb.job())

        @routes.get('/')
        async def home(_request: web.Request):
            return web.Response(body='v6d0auth\n')

        @routes.post('/approve')
        async def approve(request: web.Request):
            try:
                cert = mycdb.approve(await request.read())
            except BadSignatureError:
                raise web.HTTPUnauthorized
            except KeyError:
                raise web.HTTPNotFound
            else:
                return web.Response(body=cert)

        @routes.post('/push')
        async def push(request: web.Request):
            try:
                timeout = mycdb.push(VerifyKey(await request.read())).timeout
            except KeyError:
                raise web.HTTPTooManyRequests
            else:
                return web.Response(body=str(timeout))

        @routes.post('/pull')
        async def pull(request: web.Request):
            try:
                cert = mycdb.pull(VerifyKey(await request.read()))
            except KeyError:
                raise web.HTTPNotFound
            else:
                return web.Response(body=cert)

        @routes.get('/pullws')
        async def pullws(request: web.Request):
            ws = web.WebSocketResponse()
            await ws.prepare(request)
            try:
                srq = mycdb.requester_mapping[VerifyKey(await ws.receive_bytes())]
            except TypeError:
                await ws.close(code=http_websocket.WSCloseCode.POLICY_VIOLATION)
            else:
                try:
                    cert = await srq.future
                except asyncio.CancelledError:
                    if not srq.future.cancelled():
                        srq.future.cancel()
                    if not srq.cancelled:
                        srq.future = asyncio.get_event_loop().create_future()
                    await ws.close(code=http_websocket.WSCloseCode.TRY_AGAIN_LATER)
                else:
                    await ws.send_bytes(cert)
                    await ws.close()
            return ws
